program genealogy;
{  Program developed by David A. Hurd
                        1405 Cottage Street SW
                        Vienna, Virginia 22180
                        (703) 573-6855

     The purpose of this program was to experiment with Turbo Pascal to see
     how difficult it would be to develop a program to read the data files
     of PAF version 2 and produce reports from the data.

     The program code shows how to read and decode the fields of the INDIV2.DAT
     MARR2.DAT and NAME2.DAT files.  The name file is fairly straightforward
     if you are only retrieving names as I have done.  To update this file
     requires traversing the binary tree as described in the PAF developers
     document.

     The pascal program reads the indiv2.dat and marr2.dat files as arrays of
     bytes and then the procedures UNPACK and UNP_MARR perform the necessary
     translation of the coded information into strings and integers which
     can be used for printing.  I believe that the code is accurate and most
     of it is required to properly use the data.

     The output of this program is a Descent Chart similar to the one in FR.EXE
     but differing in that it can print an unlimited number of generations
     and will do it in compressed form.  It also shifts the print to the right
     so margins exist.  The printer codes are for EPSON FX printers, but are
     easily changed.

     A description of the function of each segment of the program is included
     with the code.
{}
{$Igdefs.pas }
{        Global File Definitions               }
var
     INDIV2 : file of indiv;
     NAME2 : file of dict;
     MARR2 : file of marr;
{}
FUNCTION FLIP(a1,a2:byte):integer;
{Reverse the bytes of integer values which are stored lo,hi by PAF
Arguments - two bytes. Returns and integer. }

   begin
      FLIP := a2*256+a1;
   end;  { FLIP }
{}
FUNCTION MNTH(mo:integer) : fmo;
{Return month names - not completely coded here for all month codes.
Argument - month code - Returns 5 character string. }

   begin
      case mo of
        1: Mnth := ' JAN ';    2: Mnth := ' FEB ';    3: Mnth := ' MAR ';
        4: Mnth := ' APR ';    5: Mnth := ' MAY ';    6: Mnth := ' JUN ';
        7: Mnth := ' JUL ';    8: Mnth := ' AUG ';    9: Mnth := ' SEP ';
       10: Mnth := ' OCT ';    11: Mnth := ' NOV ';   12: Mnth := ' DEC ';
       13: Mnth := 'NOTES';
      else Mnth := ' UNK ';
      end; { CASE }
   end; { MNTH }
{}
PROCEDURE GN(yx:integer;var leng:integer;var nam:fdate);
{Return a name string and its length from the NAME2.DAT file.  The length is
 important because this string must be concatenated character by character
 up to the length.  If this is not done, extraneous data from your NAME2
 file will appear in your output.
Arguments - NAME2.DAT pointer.  Output - String and its length. }

   var n : dict;
       i : integer;
   begin
      nam := '                ';
      SEEK(NAME2,yx);
      READ(NAME2,n);
      if yx > 0 then
      begin
         nam := n.name; leng := pos(chr(0),nam) - 1;
      end else
      begin
         nam := ''; leng := -1;
      end;
   end; { GN }
{}
PROCEDURE XDATE(a1,a2,a3,a4:byte;var res:fdate);
{Extracts a date field into a printable string.
Arguments - four bytes of packed date. Output - date string. }

   var yr,mo,day,mfr : integer;   x : string[4];    y,z : string[2];
   begin
      yr := a1*16+ a2 div 16;
      mo := (a2-(a2 div 16)*16)*2 + a3 div 128;
      day := (a3-(a3 div 128)*128) div 4;
      mfr := a3-(a3 div 4)*4;
      str(yr:4,x); str(day:2,z);
      case mfr of
        0: if (yr > 0) or (mo > 0) or (day > 0) then
             res := 'BEF '
           else
             res := '';
        1: res := 'ABT ';
        2: res := '';
        3: res := 'AFT ';
      end;
      if day > 0 then
         res := res+z;
      if mo > 0 then
         res := res+MNTH(mo);
      if yr > 0 then
         res := res+x;
      if a4 <> 0 then
      begin
        yr := yr + a4; str(yr:4,x); res := res+'/'+x;
      end;
   end; { XDATE }
{}
PROCEDURE UNP_MARR(marriage:marr;var x:mar);
{Extract Marriage file information into useable form.
Arguments - marriage record in bytes. Output - marriage record expanded.}

  var i : integer;
  begin
    with marriage,x do
    begin
      husb := FLIP(data[1],data[2]);
      wife := FLIP(data[3],data[4]);
      child := FLIP(data[5],data[6]);
      XDATE(data[7],data[8],data[9],data[10],mardate);
      mplace1 := FLIP(data[11],data[12]);
      mplace2 := FLIP(data[13],data[14]);
      mplace3 := FLIP(data[15],data[16]);
      mplace4 := FLIP(data[17],data[18]);
      XDATE(data[19],data[20],data[21],0,sealdate);
      sealtemp := FLIP(data[22],data[23]);
      hoth := FLIP(data[24],data[25]);
      woth := FLIP(data[26],data[27]);
      divflg := chr(data[28]);
    end; { UNP_MARR }
  end;
{}
PROCEDURE UNPACK(pers:indiv;var x:irec);
{Extract Individual file information into useable form.
Arguments - Individual record in bytes. Output - Individual record expanded.}

  var i : integer;
  begin
   with pers,x do
   begin
    surname := FLIP(data[1],data[2]); given1 := FLIP(data[3],data[4]);
    given2 := FLIP(data[5],data[6]);  given3 := FLIP(data[7],data[8]);
    title := FLIP(data[9],data[10]);  sex := chr(data[11]);
    XDATE(data[12],data[13],data[14],data[15],bdate);
    bplace1 := FLIP(data[16],data[17]); bplace2 := FLIP(data[18],data[19]);
    bplace3 := FLIP(data[20],data[21]); bplace4 := FLIP(data[22],data[23]);
    XDATE(data[24],data[25],data[26],data[27],cdate);
    cplace1 := FLIP(data[28],data[29]); cplace2 := FLIP(data[30],data[31]);
    cplace3 := FLIP(data[32],data[33]); cplace4 := FLIP(data[34],data[35]);
    XDATE(data[36],data[37],data[38],data[39],ddate);
    dplace1 := FLIP(data[40],data[41]); dplace2 := FLIP(data[42],data[43]);
    dplace3 := FLIP(data[44],data[45]); dplace4 := FLIP(data[46],data[47]);
    XDATE(data[48],data[49],data[50],data[51],budate);
    buplace1 := FLIP(data[52],data[53]); buplace2 := FLIP(data[54],data[55]);
    buplace3 := FLIP(data[56],data[57]); buplace4 := FLIP(data[58],data[59]);
    XDATE(data[60],data[61],data[62],0,bapdate);
    baptemp := FLIP(data[63],data[64]);
    XDATE(data[65],data[66],data[67],0,endowdate);
    endowtemp := FLIP(data[68],data[69]);
    XDATE(data[70],data[71],data[72],0,sealdate);
    sealtemp := FLIP(data[73],data[74]);
    sib := FLIP(data[75],data[76]); marr := FLIP(data[77],data[78]);
    pmarr := FLIP(data[79],data[80]);
    for i := 1 to 10 do
      id[i] := chr(data[80+i]);
    note := FLIP(data[91],data[92]);
   end;
  end; { UNPACK }
{}
PROCEDURE BLD_LINE(j:integer;var lin:fline);
{Procedure to build the line of information about an individual which will
 be printed in the Descent Chart.
Arguments - An integer RIN (j is the variable).  Output - a line of
information <name,title,RIN,birthdate,birthplace> }

   var xx : fdate;          ind : irec;          person : indiv;
   i,len : integer;         y : string[1];       yy : string[2];
   yyy : string[3];         yyyy : string[4];    yyyyy : string[5];
   begin
      SEEK(INDIV2,j); READ(INDIV2,person); UNPACK(person,ind);
      with ind do
      begin
        lin := '';
        GN(given1,len,xx);
        if len > 0 then begin
          for i := 1 to len do
             lin := lin+xx[i];
          lin := lin+' '; end;
        GN(given2,len,xx);
        if len > 0 then begin
          for i := 1 to len do
             lin := lin+xx[i];
          lin := lin+' '; end;
        GN(given3,len,xx);
        if len > 0 then begin
          for i := 1 to len do
            lin := lin+xx[i];
          lin := lin+' '; end;
        GN(surname,len,xx);
        for i := 1 to len do
          lin := lin+xx[i];
        GN(title,len,xx);
        if len > 0 then begin
          lin := lin +',';
          for i := 1 to len do
             lin := lin+xx[i]; end;
        if j>9999 then begin
          str(j:5,yyyyy); lin := lin+'('+yyyyy+')'; end
        else
          if j>999 then begin
            str(j:4,yyyy); lin := lin+'('+yyyy+')'; end
          else
            if j>99 then begin
              str(j:3,yyy); lin := lin+'('+yyy+')'; end
            else
              if j>9 then begin
                str(j:2,yy); lin := lin+'('+yy+')'; end
              else begin
                str(j:1,y); lin := lin+'('+y+')'; end;
        if length(bdate) > 3 then begin
           lin := lin + '-b.'+bdate;
           GN(bplace3,len,xx);
           if len > 0 then
              lin := lin + ','+xx; end
        else
           if length(cdate) > 3 then begin
              lin := lin + '-bap.'+cdate;
              GN(cplace3,len,xx);
              if len > 0 then
                 lin := lin + ','+xx; end
           else begin
              GN(bplace3,len,xx);
              if len > 0 then
                 lin := lin + '-b.'+xx
              else begin
                 GN(cplace3,len,xx);
                 if len > 0 then
                    lin := lin + '-bap.'+xx;
              end;
           end;
     end;
   end; { BLD_LINE }
{}
PROCEDURE DESCEND_TOP;
{Top level of the Descent Chart Process. }

   var i,rin,no_gen,Lines_Per_Page,page_no,Line_ct : integer;
       lin,hdg : fline;            ans : char;
       Print_On : boolean;         marg : string[8];
{}
   PROCEDURE HEADING;
   { Produce a heading at the top of the page }
      begin
         WRITELN(lst);WRITELN(lst);WRITELN(lst,chr(12),marg,hdg);
         WRITELN(lst,marg,'Page No:',page_no:4);WRITELN(lst);
         page_no := succ(page_no); line_ct := 1;
      end; { HEADING }
{}
   PROCEDURE DESCEND(rin,gen:integer);
   {Recursive Procedure to traverse the family descent and print the
    lines of the Descent Chart.  This routine must be studied carefully
    to understand.  The basic functions are:
                    1. Print the individual referred to by RIN.
                    2. Build a list of all marriages of that person.
                    3. For each marriage:
                       a. Print the spouse - if any.
                       b. Build a list of all children of the marriage.
                       c. For each child (in order of birth):
                          1) Recurse down one generation.
   Arguments - RIN, generation #.  Output - the report. }

      var i,j,k,m,curr_mar,curr_child : integer;
         yy : string[2];              ind,ind1 : indiv;
         pers,pers1 : irec;           m1 : marr;
         m2 : mar;                    p : genptr;
         g : genrec;                  done : boolean;
      begin
       if gen <= No_Gen then begin
         str(gen:2,yy);
         BLD_LINE(rin,lin);
         if Print_On then WRITE(lst,marg);
         for i := 1 to gen - 1 do begin
           WRITE('   '); if Print_On then WRITE(lst,'   '); end;
         WRITELN(yy:2,'-',lin);
         if Print_On then
         begin
           WRITELN(lst,yy:2,'-',lin);
           line_ct := succ(line_ct);
           if Line_ct > Lines_Per_Page then heading;
         end;
         SEEK(INDIV2,rin); READ(INDIV2,ind); UNPACK(ind,pers);
         if pers.marr <> 0 then
         begin
           new(p);
           p^.marptr := 0; p^.chptr := 0; done := False; m := pers.marr; i := 1;
           while not done do
           begin
             SEEK(MARR2,m); READ(MARR2,m1); UNP_MARR(m1,m2);
             p^.marptr := i; p^.mar[i] := m;
             if pers.sex = 'M' then
               if m2.hoth <> 0 then
               begin
                 m := m2.hoth; i := i+1;
               end else
                 done := True
             else
               if m2.woth <> 0 then
               begin
                 m := m2.woth; i := i+1;
               end else
                 done := True;
           end;
           for j := 1 to p^.marptr do
           begin
             curr_mar := p^.mar[j];
             SEEK(MARR2,curr_mar); READ(MARR2,m1); UNP_MARR(m1,m2);
             if pers.sex = 'M' then begin
               if m2.wife <> 0 then
               begin
                 BLD_LINE(m2.wife,lin);
                 if Print_On then WRITE(lst,marg);
                 for i:= 1 to gen-1 do begin
                   WRITE('   '); if Print_On then WRITE(lst,'   '); end;
                 WRITELN(' s-',lin);
                 if Print_On then
                 begin
                   WRITELN(lst,' s-',lin);
                   line_ct := succ(line_ct);
                   if line_ct > Lines_Per_Page then heading;
                 end;
               end; end
             else
               if m2.husb <> 0 then
               begin
                 BLD_LINE(m2.husb,lin);
                 if Print_On then WRITE(lst,marg);
                 for i := 1 to gen-1 do begin
                   WRITE('   '); if Print_On then WRITE(lst,'   '); end;
                 WRITELN(' s-',lin);
                 if Print_On then
                 begin
                   WRITELN(lst,' s-',lin);
                   line_ct := succ(line_ct);
                   if line_ct > Lines_Per_Page then heading;
                 end;
               end;
             if m2.child <> 0 then
             begin
               p^.chptr := 1; done := False; k := m2.child;
               p^.child[p^.chptr] := k;
               while not done do
               begin
                 SEEK(INDIV2,k); READ(INDIV2,ind1); UNPACK(ind1,pers1);
                 if pers1.sib <> 0 then
                 begin
                   p^.chptr := p^.chptr + 1; k := pers1.sib;
                   p^.child[p^.chptr] := k;
                 end else
                   done := True;
               end;
             end;
             for i := p^.chptr downto 1 do
             begin
               curr_child := p^.child[i];
               DESCEND(curr_child,gen+1);
             end;
             p^.chptr := 0;
           end;
         end;
         mark(p);
       end;
    end; { DESCEND }
{}
{Main body of DESCEND_TOP}
begin
   ClrScr; GOTOXY(10,5); WRITE('DESCENDENTS CHART');
   GOTOXY(5,8); WRITE('Enter the RIN: ');READ(rin);
   GOTOXY(5,10); WRITE('Enter the NO of Generations: ');READ(no_gen);
   GOTOXY(5,14); WRITE('Do you want to use the printer [Y or N]: ');
   READ(ans);
   BLD_LINE(rin,lin);
   hdg := '      Descent Chart for '+lin;
   if (ans = 'Y') or (ans = 'y') then
   begin
     GOTOXY(5,15); WRITE('Compressed or Normal Print? [C or N]: ');
     READ(ans);
     Print_On := True;
     page_no := 1;
     if (ans = 'C') or (ans = 'c') then
     begin
        WRITE(lst,chr(27)+'0',chr(15));
        Lines_Per_Page := 78;
        marg := '        ';
     end else
     begin
        WRITE(lst,chr(27)+'2',chr(27)+'M');
        Lines_Per_Page := 56;
        marg := '    ';
     end;
     heading;
   end else
     Print_On := False;
   GOTOXY(1,20);CLRSCR;WRITELN(no_gen:5,' Generation Chart for ',lin);
   DESCEND(rin,1);
   if Print_On then WRITE(lst,chr(12),chr(27)+'@');
end; { DESCEND_TOP }
{}
{ MAIN PROGRAM }
var i,j : integer;      Ans : char;
    lin : fline;        Done : boolean;
Begin
   ASSIGN(NAME2,'NAME2.DAT');
   RESET(NAME2);
   ASSIGN(INDIV2,'INDIV2.DAT');
   RESET(INDIV2);
   ASSIGN(MARR2,'MARR2.DAT');
   RESET(MARR2);
   done := false;
   while not done do
   begin
      CLRSCR; GOTOXY(5,10);
      WRITE('Enter Q<CR> to Quit - <CR> to Make Chart: ');
      READ(Ans);
      if (Ans <> 'Q') and (Ans <> 'q') then
      begin
         DESCEND_TOP;
         WRITELN(chr(27)+'p','Enter <CR> to continue',chr(27)+'q');
         read(ans);
      end
      else
         done := true;
   end;
end.

